<?php
 GUIFactory::setMainGUI( "shell" );
 node_class( "processlogger" ); 

 
 function &findNextDueJobs( &$jobs, &$jobRunTable, $previousJobs )
 {
 global $scheduleTypeClasses, $MAX_SCHEDULE_RUNTIME;
 $ret = array();
 
 foreach ( $jobs as $key=>$job )
 {
 $status = getJobStatus( $jobs[$key], $jobRunTable, $previousJobs );
 if ( $status == JOBSTATUS_DUE )
 $ret[] = &$jobs[$key];
 }
 return $ret;
 }

 
 function isAnyJobRunning( &$jobs, &$rootLogger , $resetbroken = FALSE ) {
 global $MAX_SCHEDULE_RUNTIME;

 
 
 if (is_null($jobs)) {
 $jobs = array();
 }

 $resetCount = 0;
 $nonParallelCount = 0;
 foreach ( $jobs as $key => $job ) {
 
 $logger = new ProcessLogger( "scheduler_task_" . $job->getValue( "id" ), true );
 if ( $logger->isRunning( $MAX_SCHEDULE_RUNTIME ) ) {
 
 if ( $job->getValue( "parallel" ) ) {
 continue;
 }
 if ( $resetbroken ) {
 $nonParallelCount++;
 } else {
 return 1;
 }
 } else {
 
 if ( $resetbroken && (!$logger->syncPid || (($pid = $logger->isRunningGetProcessId(NULL)) !== false && $pid !== 0 ))) {
 
 
 
 
 $sql = "UPDATE jobrun SET valid = -1 WHERE valid = 1 AND endtime = 0 AND job_id = '" . sqlf_escaped($job->getValue( "id" )) . "';";
 $rs = query( $sql );
 $resetCount += $rs["num"];
 if( $rs["num"] ) {
 $rootLogger->addText( "resetted broken job: " . $job->getValue( "id" ) . " Name: " . $job->getValue( "name" ), true );
 }
 }
 }
 }
 if ( $resetbroken ) {
 $rootLogger->addText( "resetting broken jobs ($resetCount)", true );
 }
 return $nonParallelCount;
 }

 
 function isJobRunning( $job_id ) {
 global $MAX_SCHEDULE_RUNTIME;
 $logger = new ProcessLogger( "scheduler_task_" . $job_id, true );
 if( ! $logger->isRunning( $MAX_SCHEDULE_RUNTIME ) ) {
 return $logger;
 }
 return FALSE;
 }
 
 
 function runTask( $taskId, $jobId, &$logger )
 {
 $cmd = translateCommand($taskId);

 
 if (preg_match("/.node\/sh.php\\s--no-owner/", $cmd)) {
 $cmd = preg_replace("/(.node\/sh.php\\s--no-owner)/", "\\1 RUNNING_JOB_ID=$jobId", $cmd);
 } else if (preg_match("/.node\/sh.php/", $cmd)) {
 $cmd = preg_replace("/(.node\/sh.php)/", "\\1 RUNNING_JOB_ID=$jobId", $cmd);
 }

 $logger->addText( "executing: $cmd", true );
 exec($cmd,$output,$return);
 $arr["output"] = implode("\n",$output);
 $arr["return"] = $return;
 $arr["cmd"] = $cmd;
 return $arr;
 }
 
 function runJob( &$job, &$jobRunTable, &$logger )
 {
 global $scheduleTypeClasses, $FEATURE;
 $jobRun = new SqlTableRow( $jobRunTable );
 $jobRun->setValue( "job_id", $job->getValue("id") );
 $jobRun->setValue( "starttime", time() );
 $jobRun->setValue( "endtime", 0 );
 $jobRun->setValue( "valid", 1 );
 $jobRun->save();
 
 if ($FEATURE["taskrunlog"]) {
 error_log("marked job " . $job->getValue("id") . " as running in database\n" , 3, $FEATURE["taskrunlog"]);
 }
 
 
 $schedule = $scheduleTypeClasses[$job->getValue( "schedule_type" )];
 $schedule->loadData( node_unserialize($job->getValue("schedule_data")) );
 $schedule->start();
 
 $return = runTask( $job->getValue( "task_id" ), $job->getValue("id"), $logger );
 
 $ret = $return["return"];
 $out = $return["output"];
 $schedule->finished();
 
 
 $jobTable = new SqlTable( "job" );
 $jobres = $jobTable->findBy( "id" , $job->getValue( "id" ));
 $job2 = &$jobres[0]; 
 if ($job2 && ($job2->getValue( "schedule_data" ) != $job->getValue( "schedule_data" ) ||
 $job2->getValue( "schedule_type" ) != $job->getValue( "schedule_type" ))) {
 $schedule = $scheduleTypeClasses[$job2->getValue( "schedule_type" )];
 $schedule->loadData( node_unserialize($job2->getValue("schedule_data")) );
 $schedule->start();
 $schedule->finished();
 $job2->setValue( "schedule_data", node_serialize( $schedule->getData() ) );
 $job2->setValue( "last_valid_jobrun_id", $jobRun->getValue( 'id' ) );
 $job2->save();
 } else {
 $job->setValue( "schedule_data", node_serialize( $schedule->getData() ) );
 $job->setValue( "last_valid_jobrun_id", $jobRun->getValue( 'id' ) );
 $job->save();
 }

 $jobRun->setValue( "endtime", time() );
 $jobRun->setValue( "returnvalue", is_numeric($ret)?$ret:-1 );
 $jobRun->setValue( "output", $out);
 $jobRun->save();

 
 $sql = "select count(*) as count, round(avg(endtime-starttime)) as average from jobrun where job_id = " . sqlf_tl_num($job->getValue( 'id' )) . " and starttime > 0 and endtime > 0 and valid = 1";
 $rs = query( $sql );
 $avgcache = $rs['arr'][0];
 $sql = "update job set jobruncount = '".sqlf_escaped($avgcache['count'])."', jobrunaverage = '".sqlf_escaped($avgcache['average']) . "' where id = ".sqlf_tl_num($job->getValue( 'id' ));
 query( $sql );

 if( $ret ) {
 
 sendFailureEmail( $return["cmd"], $job, $jobRun, $logger );
 }
 
 return $ret;
 }

 function sendFailureEmail( $cmd, &$job, &$jobRun, &$logger )
 {
 global $SCHEDULER_JOB_FAILURE_EMAIL;

 
 
 
 
 

 $from = $SCHEDULER_JOB_FAILURE_EMAIL["from"];
 $failedemails = $SCHEDULER_JOB_FAILURE_EMAIL["to"];
 $failedemails = ($failedemails ? $failedemails.',' : '' ).$job->getValue( "failedemails" );
 if( !$failedemails ) return;
 $subject = $SCHEDULER_JOB_FAILURE_EMAIL["subject"];
 if( !$subject ) $subject = "Scheduler Run Failed: #name#";
 $subject = replaceValuesForEmail( $subject, $cmd, $job, $jobRun );

 $body = $SCHEDULER_JOB_FAILURE_EMAIL['body'];
 if( !$body ) $body = "Scheduler Run Failed '#name#'\nCommand: #cmd#\nReturnValue: #returnvalue#\nStartTime: #starttime#\nEndTime: #endtime#\nOutput:\n#output#";
 $body = replaceValuesForEmail( $body, $cmd, $job, $jobRun );

 mail( $failedemails, $subject, $body, ($from ? "From: $from" : '' ) );
 }
 function replaceValuesForEmail( $str, $cmd, &$job, &$jobRun )
 {
 global $SCHEDULER_JOB_FAILURE_EMAIL;
 $dateformat = $SCHEDULER_JOB_FAILURE_EMAIL['dateformat'];
 if(!$dateformat) $dateformat = 'Y-m-d H:i:s';

 $str = str_replace( '#name#', $job->getValue( 'name' ), $str );
 $str = str_replace( '#returnvalue#', $jobRun->getValue( 'returnvalue' ), $str );
 $str = str_replace( '#output#', $jobRun->getValue( 'output' ), $str );
 $str = str_replace( '#starttime#', date( $dateformat, $jobRun->getValue( 'starttime' ) ), $str );
 $str = str_replace( '#endtime#', date( $dateformat, $jobRun->getValue( 'endtime' ) ), $str );
 $str = str_replace( '#cmd#', $cmd, $str );

 return $str;
 }

 
 
 function do_runjob( &$job, &$jobTable, &$jobRunTable, &$logger )
 {
 global $scheduleTypeClasses, $FEATURE;
 
 $ret = runJob( $job, $jobRunTable, $logger );
 $previousJobs = array();
 $previousJobs[] = $job->getValue( "id" );

 $logger->addText( "Job done, finding next due jobs.", true );
 
 
 
 
 $jobs = $jobTable->findBy( false, false, "=",  "last_valid_jobrun_id" );
 $nextJobs = &findNextDueJobs( $jobs, $jobRunTable, $previousJobs );
 $spawnednonparallel = FALSE;
 
 
 foreach ( $nextJobs as $nextJobKey => $nextJob ) {
 if ($FEATURE["taskrunlog"]) {
 error_log("finished " . $job->getValue("id") . ", followup job " . $nextJob->getValue("id"). "\n" , 3, $FEATURE["taskrunlog"]);
 }
 unset( $njob );
 unset( $schedule );
 $njob = &$nextJobs[$nextJobKey];

 if ( $njob->getValue("schedule_type") == "followup" ) {
 $scheduleType = $njob->getValue( "schedule_type" );
 $schedule = &$scheduleTypeClasses[$scheduleType];
 $schedule->loadData( node_unserialize( $njob->getValue( "schedule_data" ) ) );
 
 
 
 if ( 0 == $ret || ! $schedule->doesFollowupOnSuccess() ) {
 $schedule->d_runNext = true;
 $njob->setValue( "schedule_data", node_serialize($schedule->getData()) );
 $njob->save();
 }
 }
 }
 $logger->addText( "Spawning new scheduler process.", true );
 spawnNewScheduler();
 }

 
 
 function do_runjobs( &$nextJobs, $anyJobRunning, $spawnednonparallel = FALSE, $schedule_type = FALSE ) {
 global $FEATURE;
 $spawnedJobs = array();

 foreach ( $nextJobs as $nextJobKey => $nextJob )
 {
 if ( $schedule_type === FALSE || $nextJob->getValue("schedule_type") == $schedule_type )
 {
 $job = &$nextJobs[$nextJobKey];
 if (!Scheduler::allowRun($job->getValue("id"))) {
 continue;
 }
 if ($FEATURE["taskrunlog"]) {
 error_log("\n" . $job->getValue("name") . "(" . $job->getValue("id") . ") ", 3, $FEATURE["taskrunlog"]);
 }
 
 if ($FEATURE["advanced_parallel"]) {
 
 
 
 if (checkAncestorJobsRunning($job, $spawnedJobs) ||
 checkFollowupJobsRunningOrDue($job, $spawnedJobs) ||
 getJobStatus($job) === JOBSTATUS_RUNNING) {
 continue;
 }
 }
 if ( !$job->getValue( "parallel" ) ) {
 
 
 if( $spawnednonparallel || $anyJobRunning ) {
 continue;
 }

 $spawnednonparallel = TRUE;
 }
 
 if ($FEATURE["taskrunlog"]) {
 error_log(" STARTING " . $job->getValue("id") . "\n" , 3, $FEATURE["taskrunlog"]);
 }
 
 if ($FEATURE["advanced_parallel"]) {
 $spawnedJobs[] = $job->getValue("id");
 }
 
 spawnJobRunner( $job );
 }
 }
 
 return $spawnednonparallel;
 }

 






 function checkFollowupJobsFinished($job) {
 $followupJobIds = getFollowupJobids($job->getValue("id"));
 if (count($followupJobIds) > 0) {
 
 
 $res = query("SELECT MAX(endtime) 'endtime' FROM jobrun WHERE job_id = " . sqlf_tl_num($job->getValue("id")));
 if ($res['num'] == 0) {
 
 
 return true; 
 }
 $sql = 'SELECT jobrun.job_id FROM jobrun ' .
 'LEFT JOIN job ON job.id = jobrun.job_id ' .
 'WHERE jobrun.endtime > ' . sqlf_value($res['arr'][0]['endtime']) . ' ' . 
 'AND jobrun.job_id IN (' . implode(',', sqlf_value_array($followupJobIds)). ') ' .
 'AND job.status = 1';
 $res = query($sql);
 if ($res['num'] != count($followupJobIds)) {
 
 return false;
 }
 }
 return true;
 }
 
 




 function getFollowupJobids($jobId) {
 $followupJobIds = array();
 $res = query("SELECT jobid FROM jobdependency WHERE depid = " . sqlf_tl_num($jobId));
 if ($res['num'] > 0) {
 while (list($k,$job) = each($res['arr'])) {
 $followupJobIds[] = $job['jobid'];
 $childFollowupJobIds = getFollowupJobids($job['jobid']);
 if (count($childFollowupJobIds) > 0) {
 $followupJobIds = array_merge($followupJobIds, $childFollowupJobIds);
 }
 }
 }
 return array_unique($followupJobIds);
 }

 





 function checkFollowupJobsRunningOrDue($job, $spawnedJobs) {
 global $FEATURE;
 
 $jobId = $job->getValue("id");
 $res = query(
"SELECT j.id, j.status " .
 "FROM job j " .
 "LEFT JOIN jobdependency jd " .
 "ON jd.jobid = j.id " .
 "WHERE jd.depid = ".sqlf_tl_num($jobId));
 if ($res['num'] > 0) {
 $jobTable = new SqlTable( "job" );
 foreach ($res['arr'] as $k => $jobRow) {
 if (in_array($jobRow["id"], $spawnedJobs)) {
 if ($FEATURE["taskrunlog"]) {
 error_log("a[" . $jobRow["id"] . ">S]" , 3, $FEATURE["taskrunlog"]);
 }
 return true;
 }
 $followupJob = $jobTable->findBy( "id", $jobRow["id"] );
 $followupJob = $followupJob[0];
 if ($FEATURE["taskrunlog"]) {
 error_log("f[" . $followupJob->getValue("id") . ">" . getJobStatus($followupJob) . "]" , 3, $FEATURE["taskrunlog"]);
 }
 if (getJobStatus($followupJob) === JOBSTATUS_RUNNING ||
 getJobStatus($followupJob) === JOBSTATUS_DUE ||
 checkFollowupJobsRunningOrDue($followupJob, $spawnedJobs)) {
 return true;
 }
 }
 }
 return false;
 }
 
 





 function checkAncestorJobsRunning($job, $spawnedJobs) {
 global $FEATURE;
 
 $jobId = $job->getValue("id");
 $res = query(
"SELECT jd.depid, j.status " .
 "FROM job j " .
 "LEFT JOIN jobdependency jd " .
 "ON jd.depid = j.id " .
 "WHERE jd.jobid = ".sqlf_tl_num($jobId));
 if ($res['num'] > 0) {
 $jobTable = new SqlTable( "job" );
 foreach ($res['arr'] as $k => $jobRow) {
 if (in_array($jobRow["depid"], $spawnedJobs)) {
 if ($FEATURE["taskrunlog"]) {
 error_log("a[" . $jobRow["depid"] . ">S]" , 3, $FEATURE["taskrunlog"]);
 }
 return true;
 }
 $ancestorJob = $jobTable->findBy( "id", $jobRow["depid"] );
 $ancestorJob = $ancestorJob[0];
 if ($FEATURE["taskrunlog"]) {
 error_log("a[" . $ancestorJob->getValue("id") . ">" . getJobStatus($ancestorJob) . "]" , 3, $FEATURE["taskrunlog"]);
 }
 if (getJobStatus($ancestorJob) === JOBSTATUS_RUNNING ||
 checkAncestorJobsRunning($ancestorJob, $spawnedJobs)) {
 return true;
 }
 }
 }
 return false; 
 }

 
 function spawnJobRunner( &$job )
 {
 spawnNewScheduler( "runjob", $job->getValue( "id" ) );
 }


 $HTML_HEADER = false;
 $HTML_FOOTER = false;
 $HTTP_HEADER = false;
 
 if( $cmd != "runjob" ) {
 $logger = new ProcessLogger( "scheduler", true );
 } else {
 $logger = new ProcessLogger( "scheduler_task_" . $job_id );
 }

 
 if ( !$logger->checkRunningAndStart( $MAX_SCHEDULE_RUNTIME ) )
 {
 $logger->skip();
 return;
 }
 
 

 $logger->addText( "", true );

 node_include( "task_schedule_func" );
 node_include( "scheduletype" );
 node_class( "sqltable" );
 
 $jobRunTable = new SqlTable( "jobrun" );
 $jobTable = new SqlTable( "job" );
 global $SETTINGS;
 
 node_class("scheduler");
 
 if( $cmd == "runjob" && Scheduler::allowRun($job_id)) {
 
 $jobres = $jobTable->findBy( "id", $job_id );
 $job = &$jobres[0];
 
 if( ! $job->getValue( "parallel") && isAnyJobRunning( $jobs, $logger ) ) {
 $logger->skip();
 } else {
 do_runjob( $job, $jobTable, $jobRunTable, $logger );
 $logger->stop();
 }
 } else {
 
 
 
 
 $jobs = $jobTable->findBy( false, false, "=",  "last_valid_jobrun_id" );

 $logger->addText( "checking for due jobs...", true );

 $anyJobRunning = isAnyJobRunning( $jobs, $logger, TRUE );

 $logger->addText( "anyJobRunning?: $anyJobRunning", true );
 
 $previousJobs = array();
 
 $firstJob = TRUE;
 
 $nextJobs = &findNextDueJobs( $jobs, $jobRunTable, $previousJobs );

 $logger->addText( "Running followup jobs..." );
 $spawnednonparallel = do_runjobs( $nextJobs, $anyJobRunning, FALSE, "followup" ); 
 $logger->addText( "Running all other jobs..." );
 $spawnednonparallel = do_runjobs( $nextJobs, $anyJobRunning, $spawnednonparallel );
 
 $logger->stop();
 }
?>